from ._base import BaseImage


class Kafka(BaseImage):
    label = 'kafka'
    name = 'kafka'
    port = 9092

    def check(self):
        from kafka import KafkaClient
        from kafka.common import KafkaUnavailableError
        try:
            KafkaClient(f"{self.host}:{self.get_port()}")
            return True
        except KafkaUnavailableError:
            pass
        return False


kafka_image = Kafka()



from pprint import pformat
from pytest_docker_fixtures import images
from time import sleep

import docker
import os


class BaseImage:

    docker_version = '1.23'
    name = 'foobar'
    port = None
    host = ''
    base_image_options = dict(
        cap_add=['IPC_LOCK'],
        mem_limit='1g',
        environment={},
        privileged=True,
        detach=True,
        publish_all_ports=True)

    @property
    def image(self):
        return images.get_image(self.name)

    def get_image_options(self):
        image_options = self.base_image_options.copy()
        if 'environment' not in image_options:
            image_options['environment'] = {}
        for key, value in images.get_env(self.name).items():
            if value is None:
                if key in image_options['environment']:
                    del image_options['environment'][key]
            else:
                image_options['environment'][key] = value
        image_options.update(images.get_options(self.name))
        image_options['max_wait_s'] = images.get_max_wait_s(self.name)
        return image_options

    def get_port(self, port=None):
        if (os.environ.get('TESTING', '') == 'jenkins' or
                'TRAVIS' in os.environ):
            return port if port else self.port
        network = self.container_obj.attrs['NetworkSettings']
        service_port = '{0}/tcp'.format(port if port else self.port)
        for netport in network['Ports'].keys():
            if netport == '6543/tcp':
                continue

            if netport == service_port:
                return network['Ports'][service_port][0]['HostPort']

    def get_host(self):
        return self.container_obj.attrs['NetworkSettings']['IPAddress']

    def check(self):
        return True

    def run(self):
        docker_client = docker.from_env(version=self.docker_version)
        image_options = self.get_image_options()

        max_wait_s = image_options.pop('max_wait_s', None) or 30

        # Create a new one
        container = docker_client.containers.run(
            image=self.image,
            **image_options
        )
        ident = container.id
        count = 1

        self.container_obj = docker_client.containers.get(ident)

        opened = False

        print(f'starting {self.name}')
        while count < max_wait_s and not opened:
            if count > 0:
                sleep(1)
            count += 1
            try:
                self.container_obj = docker_client.containers.get(ident)
            except docker.errors.NotFound:
                print(f'Container not found for {self.name}')
                continue
            if self.container_obj.status == 'exited':
                logs = self.container_obj.logs()
                self.stop()
                raise Exception(f'Container failed to start {logs}')

            if self.container_obj.attrs['NetworkSettings']['IPAddress'] != '':
                if os.environ.get('TESTING', '') == 'jenkins':
                    network = self.container_obj.attrs['NetworkSettings']
                    self.host = network['IPAddress']
                else:
                    self.host = 'localhost'

            if self.host != '':
                opened = self.check()
        if not opened:
            logs = self.container_obj.logs().decode('utf-8')
            self.stop()
            raise Exception(
                f'Could not start {self.name}: {logs}\n'
                f'Image: {self.image}\n'
                f'Options:\n{pformat(image_options)}')
        print(f'{self.name} started')
        return self.host, self.get_port()

    def stop(self):
        if self.container_obj is not None:
            try:
                self.container_obj.kill()
            except docker.errors.APIError:
                pass
            try:
                self.container_obj.remove(v=True, force=True)
            except docker.errors.APIError:
                pass



from ._base import BaseImage


class MySQL(BaseImage):
    name = 'mysql'
    port = 3306

    def check(self):
        import mysql.connector

        conn = None
        cur = None

        try:
            conn = mysql.connector.connect(
                host=self.host,
                port=self.get_port(),
                user="root",
                use_pure=True,
            )

            try:
                cur = conn.cursor()
                cur.execute("SELECT 1;")
                cur.fetchone()
            finally:
                if cur:
                    cur.close()

            return True
        except Exception:
            return False
        finally:
            if conn:
                conn.close()


mysql_image = MySQL()



from ._base import BaseImage


class CockroachDB(BaseImage):
    name = 'cockroach'
    port = 26257

    def check(self):
        import psycopg2
        conn = cur = None
        try:
            conn = psycopg2.connect(
                f"dbname=guillotina user=root host={self.host} "
                f"port={self.get_port()}")
            conn.set_session(autocommit=True)
            cur = conn.cursor()
            cur.execute('SHOW DATABASES;')
            for result in cur.fetchall():
                if result[0] == 'guillotina':
                    conn.close()
                    cur.close()
                    return True
            cur.execute("CREATE DATABASE IF NOT EXISTS guillotina;")
            cur.close()
            conn.close()
        except: # noqa
            if conn is not None:
                conn.close()
            if cur is not None:
                cur.close()
        return False


cockroach_image = CockroachDB()



settings = {
    'cockroach': {
        'image': 'cockroachdb/cockroach',
        'version': 'v1.1.7',
        'options': {
            'command': ' '.join([
                'start --insecure',
            ]),
            'publish_all_ports': False,
            'ports': {
                f'26257/tcp': '26257'
            }
        }
    },
    'elasticsearch': {
        'image': 'elasticsearch',
        'version': '5.2.0',
        'env': {
            'cluster.name': 'docker-cluster',
            'ES_JAVA_OPTS': '-Xms512m -Xmx512m',
            "xpack.security.enabled": "false"
        },
        'options': {
            'cap_add': ['IPC_LOCK'],
            'mem_limit': '1g'
        }
    },
    'etcd': {
        'image': 'quay.io/coreos/etcd',
        'version': 'v3.2.0-rc.0'
    },
    'minio': {
        'image': 'minio/minio',
        'version': 'latest',
        'env': {
            'MINIO_ACCESS_KEY': 'x' * 10,
            'MINIO_SECRET_KEY': 'x' * 10,
        },
        'options': {
            'cap_add': ['IPC_LOCK'],
            'mem_limit': '200m',
            'command': 'server /export',
        }
    },
    'mysql': {
        'image': 'mysql',
        'version': '5.7',
        'env': {
            'MYSQL_ALLOW_EMPTY_PASSWORD': 'yes'
        }
    },
    'postgresql': {
        'image': 'postgres',
        'version': '9.6.16',
        'env': {
            'POSTGRES_PASSWORD': '',
            'POSTGRES_DB': 'guillotina',
            'POSTGRES_USER': 'postgres'
        }
    },
    'redis': {
        'image': 'redis',
        'version': '3.2.8',
        'options': {
            'cap_add': ['IPC_LOCK'],
            'mem_limit': '200m'
        }
    },
    'rabbitmq': {
        'image': 'rabbitmq',
        'version': '3.7.8'
    },
    'kafka': {
        'image': 'spotify/kafka',
        'version': 'latest',
        'env': {
            'ADVERTISED_PORT': '9092',
            'ADVERTISED_HOST': '0.0.0.0'
        },
        'options': {
            'ports': {
                '9092': '9092',
                '2181': '2181'
            }
        }
    }
}


def get_image(name):
    image = settings[name]
    return image['image'] + ':' + image['version']


def configure(name, image=None, version=None, full=None,
              env=None, options=None, max_wait_s=None):
    if full is not None:
        image, _, version = full.partition(':')
    if image is not None:
        settings[name]['image'] = image
    if version is not None:
        settings[name]['version'] = version
    if env is not None:
        if 'env' not in settings[name]:
            settings[name]['env'] = {}
        settings[name]['env'].update(env)
    if max_wait_s:
        settings[name]['max_wait_s'] = max_wait_s


def get_env(name):
    image = settings[name]
    return image.get('env') or {}

def get_max_wait_s(name):
    # Default to 30 seconds
    image = settings[name]
    return image.get('max_wait_s') or 30

def get_options(name):
    image = settings[name]
    return image.get('options') or {}



from ._base import BaseImage


class Postgresql(BaseImage):
    name = 'postgresql'
    port = 5432

    def check(self):
        import psycopg2
        try:
            conn = psycopg2.connect(
                f"dbname=guillotina user=postgres host={self.host} "
                f"port={self.get_port()}")
            cur = conn.cursor()
            cur.execute("SELECT 1;")
            cur.fetchone()
            cur.close()
            conn.close()
            return True
        except: # noqa
            conn = None
            cur = None
        return False


pg_image = Postgresql()



from ._base import BaseImage

import requests


class Minio(BaseImage):
    name = 'minio'
    port = 9000

    def check(self):
        url = f'http://{self.host}:{self.get_port()}/'
        try:
            resp = requests.options(url)
            if 200 <= resp.status_code < 500:
                return True
        except Exception:
            pass
        return False


minio_image = Minio()



from .containers.cockroach import cockroach_image
from .containers.es import es_image
from .containers.etcd import etcd_image
from .containers.pg import pg_image
from .containers.redis import redis_image
from .containers.rabbitmq import rabbitmq_image
from .containers.kafka import kafka_image
from .containers.minio import minio_image
from .containers.mysql import mysql_image

import os
import pytest


IS_TRAVIS = 'TRAVIS' in os.environ


@pytest.fixture(scope='session')
def redis():
    """
    detect travis, use travis's postgres; otherwise, use docker
    """
    if os.environ.get('REDIS'):
        yield os.environ['REDIS'].split(':')
    else:
        if IS_TRAVIS:
            host = 'localhost'
            port = 6379
        else:
            host, port = redis_image.run()

        yield host, port  # provide the fixture value

        if not IS_TRAVIS:
            redis_image.stop()


@pytest.fixture(scope='session')
def cockroach():
    if os.environ.get('COCKROACH'):
        yield os.environ['COCKROACH'].split(':')
    else:
        yield cockroach_image.run()
        cockroach_image.stop()


@pytest.fixture(scope='session')
def pg():
    if os.environ.get('POSTGRESQL'):
        yield os.environ['POSTGRESQL'].split(':')
    else:
        if IS_TRAVIS:
            host = 'localhost'
            port = 6379
        else:
            host, port = pg_image.run()

        yield host, port  # provide the fixture value

        if not IS_TRAVIS:
            pg_image.stop()


@pytest.fixture(scope='session')
def etcd():
    if os.environ.get('ETCD'):
        yield os.environ['ETCD'].split(':')
    else:
        yield etcd_image.run()
        etcd_image.stop()


@pytest.fixture(scope='session')
def es():
    if os.environ.get('ELASTICSEARCH'):
        yield os.environ['ELASTICSEARCH'].split(':')
    else:
        yield es_image.run()
        es_image.stop()


@pytest.fixture(scope='session')
def rabbitmq():
    if os.environ.get('RABBITMQ'):
        yield os.environ['RABBITMQ'].split(':')
    else:
        yield rabbitmq_image.run()
        rabbitmq_image.stop()


@pytest.fixture(scope='session')
def kafka():
    if os.environ.get('KAFKA'):
        yield os.environ['KAFKA'].split(':')
    else:
        yield kafka_image.run()
        kafka_image.stop()


@pytest.fixture(scope='session')
def minio():
    if os.environ.get('MINIO'):
        yield os.environ['MINIO'].split(':')
    else:
        if IS_TRAVIS:
            host = 'localhost'
            port = 19000
        else:
            host, port = minio_image.run()

        yield host, port

        if not IS_TRAVIS:
            minio_image.stop()


@pytest.fixture(scope='session')
def mysql():
    if os.environ.get('MYSQL'):
        yield os.environ['MYSQL'].split(':')
    else:
        yield mysql_image.run()
        mysql_image.stop()
